@@ -1,6 +1,7 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of
 # this repository contains the full copyright notices and license terms.
-import time
+import datetime as dt
+import urllib.parse
 from decimal import Decimal
 
 import pyactiveresource
@@ -10,18 +11,18 @@ from shopify.api_version import ApiVersion
 from trytond.cache import Cache
 from trytond.config import config
 from trytond.i18n import gettext
-from trytond.model import ModelSQL, ModelView, fields, Unique
-from trytond.model import sequence_ordered, MatchMixin
-from trytond.pool import PoolMeta, Pool
+from trytond.model import (
+    MatchMixin, ModelSQL, ModelView, Unique, fields, sequence_ordered)
+from trytond.pool import Pool, PoolMeta
 from trytond.pyson import Eval
+from trytond.tools import grouped_slice
 from trytond.transaction import Transaction
+from trytond.url import http_host
 
 from .common import IdentifierMixin, IdentifiersMixin
 from .exceptions import ShopifyError
 
-BACKOFF_TIME = config.getfloat(
-    'web_shop_shopify', 'api_backoff_time', default=1)
-BACKOFF_TIME_FACTOR = 12
+EDIT_ORDER_DELAY = dt.timedelta(days=60 + 1)
 
 
 class Shop(metaclass=PoolMeta):
@@ -31,24 +32,28 @@ class Shop(metaclass=PoolMeta):
         'required': Eval('type') == 'shopify',
         'invisible': Eval('type') != 'shopify',
         }
-    _depends = ['type']
 
-    shopify_url = fields.Char("Shop URL", states=_states, depends=_depends)
+    shopify_url = fields.Char("Shop URL", states=_states)
     shopify_version = fields.Selection(
-        'get_shopify_versions', "Version", states=_states, depends=_depends)
-    shopify_trial = fields.Boolean(
-        "Trial",
-        help="If checked reduce the number of calls per minute.")
-    shopify_password = fields.Char(
-        "Password", states=_states, depends=_depends)
+        'get_shopify_versions', "Version", states=_states)
+    shopify_password = fields.Char("Password", states=_states)
+    shopify_webhook_shared_secret = fields.Char(
+        "Webhook Shared Secret",
+        states={
+            'invisible': _states['invisible'],
+            })
+    shopify_webhook_endpoint_order = fields.Function(
+        fields.Char(
+            "Webhook Order Endpoint",
+            help="The URL to be called by Shopify for Order events."),
+        'on_change_with_shopify_webhook_endpoint_order')
     shopify_warehouses = fields.One2Many(
-        'web.shop-stock.location', 'shop', "Warehouses",
-        states=_states, depends=_depends)
+        'web.shop-stock.location', 'shop', "Warehouses", states=_states)
     shopify_payment_journals = fields.One2Many(
         'web.shop.shopify_payment_journal', 'shop', "Payment Journals",
-        states=_states, depends=_depends)
+        states=_states)
 
-    del _states, _depends
+    del _states
 
     @classmethod
     def __setup__(cls):
@@ -60,13 +65,25 @@ class Shop(metaclass=PoolMeta):
                 field.states['invisible'] |= invisible
             else:
                 field.states['invisible'] = invisible
-            field.depends.append('type')
 
     @classmethod
     def get_shopify_versions(cls):
         return [(None, "")] + sorted(
             ((v, v) for v in ApiVersion.versions), reverse=True)
 
+    @fields.depends('name')
+    def on_change_with_shopify_webhook_endpoint_order(self, name=None):
+        if not self.name:
+            return
+        url_part = {
+            'database_name': Transaction().database.name,
+            'shop': self.name,
+            }
+        return http_host() + (
+            urllib.parse.quote(
+                '/%(database_name)s/web_shop_shopify/webhook/%(shop)s/order' %
+                url_part))
+
     @classmethod
     def view_attributes(cls):
         return super().view_attributes() + [
@@ -76,12 +93,13 @@ class Shop(metaclass=PoolMeta):
             ]
 
     @classmethod
-    def validate(cls, shops):
-        super().validate(shops)
-        for shop in shops:
-            if shop.type == 'shopify':
-                for product in shop.products:
-                    shop._shopify_check_product(product)
+    def validate_fields(cls, shops, field_names):
+        super().validate_fields(shops, field_names)
+        if field_names & {'type', 'products'}:
+            for shop in shops:
+                if shop.type == 'shopify':
+                    for product in shop.products:
+                        shop._shopify_check_product(product)
 
     def _shopify_check_product(self, product):
         if not product.template.shopify_uom:
@@ -115,24 +133,28 @@ class Shop(metaclass=PoolMeta):
                     and payment_journal.match(pattern)):
                 return payment_journal.journal
 
+    def managed_metafields(self):
+        return set()
+
     def __sync_metafields(self, resource, metafields):
+        metafields = metafields.copy()
+        managed_metafields = self.managed_metafields()
+        assert metafields.keys() <= managed_metafields
         for metafield in resource.metafields():
             key = '.'.join([metafield.namespace, metafield.key])
             value = metafield.to_dict()
             if key not in metafields:
-                metafield.destroy()
-                time.sleep(BACKOFF_TIME)
+                if key in managed_metafields:
+                    metafield.destroy()
             elif metafields[key] != value:
                 for k, v in metafields.pop(key).items():
                     setattr(metafield, k, v)
                 metafield.save()
-                time.sleep(BACKOFF_TIME)
         for key, value in metafields.items():
             namespace, key = key.split('.', 1)
             value['namespace'] = namespace
             value['key'] = key
             resource.add_metafield(shopify.Metafield(value))
-            time.sleep(BACKOFF_TIME)
 
     @classmethod
     def shopify_update_product(cls, shops=None):
@@ -222,7 +244,6 @@ class Shop(metaclass=PoolMeta):
             identifier.to_update = False
             identifier.save()
         Transaction().commit()
-        time.sleep(BACKOFF_TIME)
 
         self.__sync_metafields(
             custom_collection, category.get_shopify_metafields(self))
@@ -232,7 +253,6 @@ class Shop(metaclass=PoolMeta):
         if shopify_id:
             if shopify.CustomCollection.exists(shopify_id):
                 shopify.CustomCollection.find(shopify_id).destroy()
-            time.sleep(BACKOFF_TIME)
             category.set_shopify_identifier(self)
 
     def __shopify_update_template(
@@ -268,7 +288,6 @@ class Shop(metaclass=PoolMeta):
             variant, = shopify_product.variants
             product.set_shopify_identifier(self, variant.id)
         Transaction().commit()
-        time.sleep(BACKOFF_TIME)
 
         self.__sync_metafields(
             shopify_product, template.get_shopify_metafields(self))
@@ -302,7 +321,6 @@ class Shop(metaclass=PoolMeta):
             return
         if shopify.Product.exists(shopify_id):
             shopify.Product.find(shopify_id).destroy()
-        time.sleep(BACKOFF_TIME)
         template.set_shopify_identifier(self)
         for product in template.products:
             product.set_shopify_identifier(self)
@@ -335,12 +353,10 @@ class Shop(metaclass=PoolMeta):
                 image.set_shopify_identifier(
                     self, product_image.id)
                 transaction.commit()
-            time.sleep(BACKOFF_TIME)
 
         for image in shopify_product.images:
             if image.id not in image_ids:
                 image.destroy()
-                time.sleep(BACKOFF_TIME)
 
     def __shopify_update_product(self, shopify_shop, product, price, tax):
         update_extra = {'price': str(price), 'tax': str(tax)}
@@ -362,7 +378,6 @@ class Shop(metaclass=PoolMeta):
             identifier.to_update_extra = update_extra
             identifier.save()
         Transaction().commit()
-        time.sleep(BACKOFF_TIME)
 
         self.__sync_metafields(variant, product.get_shopify_metafields(self))
 
@@ -384,13 +399,11 @@ class Shop(metaclass=PoolMeta):
                 identifier.to_update = False
                 identifier.save()
             Transaction().commit()
-        time.sleep(BACKOFF_TIME)
 
     def __shopify_remove_product(self, product):
         shopify_id = product.get_shopify_identifier(self)
         if shopify_id:
             if shopify.Variant.exists(shopify_id):
-                time.sleep(BACKOFF_TIME)
                 shopify.Variant.find(shopify_id).destroy()
             product.set_shopify_identifier(self)
 
@@ -441,9 +454,6 @@ class Shop(metaclass=PoolMeta):
                         location_id, inventory_item_id, quantity)
                 except pyactiveresource.connection.ResourceNotFound:
                     pass
-                time.sleep(BACKOFF_TIME)
-            elif not i % 250:
-                time.sleep(BACKOFF_TIME)
 
         for product_id, quantity in product2quantity.items():
             inventory_item_id = product2shopify.get(product_id)
@@ -454,7 +464,6 @@ class Shop(metaclass=PoolMeta):
                     location_id, inventory_item_id, quantity)
             except pyactiveresource.connection.ResourceNotFound:
                 pass
-            time.sleep(BACKOFF_TIME)
 
     @classmethod
     def shopify_fetch_order(cls, shops=None):
@@ -489,8 +498,6 @@ class Shop(metaclass=PoolMeta):
                 sales = []
                 for i, order in enumerate(orders):
                     sales.append(Sale.get_from_shopify(shop, order))
-                    if i % 250:
-                        time.sleep(BACKOFF_TIME)
                 Sale.save(sales)
                 for sale, order in zip(sales, orders):
                     sale.shopify_tax_adjustment = (
@@ -502,26 +509,76 @@ class Shop(metaclass=PoolMeta):
 
     @classmethod
     def shopify_update_order(cls, shops=None):
-        """Update existing Shopify Order"""
+        """Update existing sale from Shopify"""
         pool = Pool()
         Sale = pool.get('sale.sale')
-        Payment = pool.get('account.payment')
         if shops is None:
             shops = cls.search([
                     ('type', '=', 'shopify'),
                     ])
         cls.lock(shops)
+        now = dt.datetime.now()
         for shop in shops:
             sales = Sale.search([
                     ('web_shop', '=', shop.id),
                     ('shopify_identifier', '!=', None),
-                    ('state', 'in', ['quotation', 'confirmed', 'processing']),
+                    ['OR',
+                        ('state', 'in',
+                            ['quotation', 'confirmed', 'processing']),
+                        ('create_date', '>=', now - EDIT_ORDER_DELAY),
+                        ],
                     ])
+            for sub_sales in grouped_slice(sales, count=250):
+                cls._shopify_update_order(shop, list(sub_sales))
+
+    @classmethod
+    def _shopify_update_order(cls, shop, sales):
+        assert shop.type == 'shopify'
+        assert all(s.web_shop == shop for s in sales)
+        with shop.shopify_session():
+            orders = shopify.Order.find(
+                ids=','.join(str(s.shopify_identifier) for s in sales),
+                status='any')
+            id2order = {o.id: o for o in orders}
+
+        to_update = []
+        orders = []
+        for sale in sales:
+            try:
+                order = id2order[sale.shopify_identifier]
+            except KeyError:
+                continue
+            to_update.append(sale)
+            orders.append(order)
+        cls.shopify_update_sale(to_update, orders)
+
+    @classmethod
+    def shopify_update_sale(cls, sales, orders):
+        """Update sales based on Shopify orders"""
+        pool = Pool()
+        Amendment = pool.get('sale.amendment')
+        Payment = pool.get('account.payment')
+        Sale = pool.get('sale.sale')
+        assert len(sales) == len(orders)
+        to_update = {}
+        for sale, order in zip(sales, orders):
+            assert sale.shopify_identifier == order.id
+            shop = sale.web_shop
             with shop.shopify_session():
-                for sale in sales:
-                    order, = shopify.Order.find(ids=sale.shopify_identifier)
-                    Payment.get_from_shopify(sale, order)
-                    time.sleep(BACKOFF_TIME)
+                sale = Sale.get_from_shopify(shop, order, sale=sale)
+                if sale._changed_values:
+                    sale.untaxed_amount_cache = None
+                    sale.tax_amount_cache = None
+                    sale.total_amount_cache = None
+                    to_update[sale] = order
+                Payment.get_from_shopify(sale, order)
+        Sale.save(to_update.keys())
+        for sale, order in to_update.items():
+            sale.shopify_tax_adjustment = (
+                Decimal(order.current_total_price) - sale.total_amount)
+        Sale.store_cache(to_update.keys())
+        Amendment._clear_sale(to_update.keys())
+        Sale.__queue__.process(to_update.keys())
 
 
 class ShopShopifyIdentifier(IdentifierMixin, ModelSQL, ModelView):
